#include <httpd.h>
#include <http_protocol.h>
#include <http_config.h>
#include <syslog.h>
#include <zlib.h>

#define BUFFER 8192


/* mod_gzip - copyright 2007 ed neville
 * reads a .gz file to the client in substitution for the .html with
 * gzip content-encoding. if the client does not accept the gzip
 * encoding then a decompressed format will be sent.
 *
 * 2006-05-14 ed AT s5h DOT net - beta code
 * Use at your own risk. Author not responsible for use.
 *
 * Updates at http://www.s5h.net
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * $URL: svn+ssh://fw/repos/code/apache2_modules/gzip/mod_gzip_disk.c $
 * $Id: mod_gzip_disk.c 99 2008-02-13 19:05:06Z ed $
 */

/**
 * this function just outputs to the syslog a key/value pair.
 */
void debug_print_chr_ptr( request_rec *r, const char *key, char *val ) {
	char *ptr = (char *)apr_palloc( r->pool, 
		strlen( key ) + strlen( val ) + 3 );

	if( ptr != NULL ) {
		sprintf( ptr, "%s: %s", key, val );
		syslog( LOG_DEBUG, ptr );
	}
	else {
		syslog( LOG_DEBUG, "not enough room to create log" );
	}
}

/**
 * this is the callback function from the apr_table_do function. we're
 * not using the generic function above since we want the prefix header:
 */
int header_do( request_rec *r, const char *key, const char *val ) {
	char *temp = (char *)apr_palloc( r->pool, strlen( key ) + 20 );
	sprintf( temp, "Header %s", key );
	debug_print_chr_ptr( r, temp, val );
	return(1);
}

/**
 * here we just debug all the request character strings and tables
 */
int print_request( request_rec *r ) {
	if( r->the_request != NULL ) {
		debug_print_chr_ptr( r, "the_request", (char *)r->the_request );
	}
	if( r->protocol != NULL ) {
		debug_print_chr_ptr( r, "protocol", (char *)r->protocol );
	}
	if( r->hostname != NULL ) {
		debug_print_chr_ptr( r, "hostname", (char *)r->hostname );
	}
	if( r->status_line != NULL ) {
		debug_print_chr_ptr( r, "status_line", (char *)r->status_line );
	}
	if( r->method != NULL ) {
		debug_print_chr_ptr( r, "method", (char *)r->method );
	}
	if( r->range != NULL ) {
		debug_print_chr_ptr( r, "range", (char *)r->range );
	}
	if( r->content_type != NULL ) {
		debug_print_chr_ptr( r, "content_type", (char *)r->content_type );
	}	/* Break these out --- we dispatch on 'em */
	if( r->handler != NULL ) {
		debug_print_chr_ptr( r, "handler", (char *)r->handler );
	}	/* What we *really* dispatch on */
	if( r->content_encoding != NULL ) {
		debug_print_chr_ptr( r, "content_encoding", (char *)r->content_encoding );
	}
	if( r->vlist_validator != NULL ) {
		debug_print_chr_ptr( r, "vlist_validator", (char *)r->vlist_validator );
	}
	if( r->user != NULL ) {
		debug_print_chr_ptr( r, "user", (char *)r->user );
	}	
	if( r->ap_auth_type != NULL ) {
		debug_print_chr_ptr( r, "ap_auth_type", (char *)r->ap_auth_type );
	}
	if( r->unparsed_uri != NULL ) {
		debug_print_chr_ptr( r, "unparsed_uri", (char *)r->unparsed_uri );
	}	
	if( r->uri != NULL ) {
		debug_print_chr_ptr( r, "uri", (char *)r->uri );
	}
	if( r->filename != NULL ) {
		debug_print_chr_ptr( r, "filename", (char *)r->filename );
	}
	if( r->canonical_filename != NULL ) {
		debug_print_chr_ptr( r, "canonical_filename", (char *)r->canonical_filename );
	}
	if( r->path_info != NULL ) {
		debug_print_chr_ptr( r, "path_info", (char *)r->path_info );
	}
	if( r->args != NULL ) {
		debug_print_chr_ptr( r, "args", (char *)r->args );
	}
//	if( r->headers_in != NULL ) {
	
	syslog( LOG_DEBUG, "HEADERS IN" );
		apr_table_do( header_do, r, 
			r->headers_in, NULL );
	syslog( LOG_DEBUG, "END HEADERS IN" );
//	}
//	if( r->headers_out != NULL ) {
	syslog( LOG_DEBUG, "HEADERS OUT" );
		apr_table_do( header_do, r, 
			r->headers_out, NULL );
	syslog( LOG_DEBUG, "END HEADERS OUT" );
//	}
//	if( r->err_headers_out != NULL ) {
	syslog( LOG_DEBUG, "HEADERS ERR OUT" );
		apr_table_do( header_do, r, 
			r->err_headers_out, NULL );
	syslog( LOG_DEBUG, "HEADERS ERR OUT" );
//	}


	return(0);
}

/**
 * this function determines if the user can accept the gzip encoding.
 */

static int user_can_accept_gzip( request_rec *r ) {
	char *ptr;
	if( ptr = apr_table_get( r->headers_in, "Accept-Encoding" ) ) {
		if( strstr( ptr, "gzip" ) != NULL ) {
			return( EXIT_SUCCESS );
		}
	}
	return( EXIT_FAILURE );
}

/*
 * it would be nice to integrate with mod_mime here, but i'm not sure how well that will go, must read the code a little more FIXME
 */
static int set_mime_type( request_rec *r, char *s ) {
	char *ptr, *p;

	p = ptr = (char *)apr_palloc( r->pool, strlen( s ) + 1 );
	
	strncpy( ptr, s, strlen(s)+1 );

	while( strchr( ptr, '.' ) ) {
		ptr++;
	}

	if( ptr > p ) {
		(--ptr)[0] = '\0';
	}

	ptr = p;

	while( strchr( ptr, '.' ) ) {
		ptr++;
	}

	/* 
	 * syslog( LOG_DEBUG, "PTR: " );
	 * syslog( LOG_DEBUG, ptr );
	 * syslog( LOG_DEBUG, p );
	 */

	if( strncmp( ptr, "html", 4 ) == 0 
		|| strncmp( ptr, "php", 3 ) == 0 ) {
		ap_set_content_type( r, "text/html;charset=ascii" );
		return( EXIT_SUCCESS );
	}

	ap_set_content_type( r, "text/plain" );

	return( EXIT_SUCCESS );
}


static int gzip_disk_header_parser( request_rec *r ) {

	struct stat s;
	apr_size_t content_sent;
	/*
	 * char *buffer = (void *)apr_palloc( r->pool, BUFFER );
	 */
	int size_of_scratch = strlen( r->filename ) + BUFFER;
	char *scratch = (char *)apr_palloc( r->pool, size_of_scratch );
	size_t count;
	gzFile *gz;

	/**
	 * this is helpful for debugging purposes, but we don't need it
	 * for day-to-day, business as usual work
	 *
	 * print_request( r );
	 */

	/**
	 * we need at least BUFFER anyway, so don't shout about BUFFER
	 * being greater than 4 (which is the most needed for the
	 * extension .gz
	 */

	/*
	 * syslog( LOG_DEBUG, __FUNCTION__ );
	 */

	if( r->method_number != M_GET ) {
		return( DECLINED );
	}

	/*
	 * syslog( LOG_DEBUG, r->filename );
	 */

	if( strlen( r->filename ) > 2 ) {
		if( strcmp( r->filename + ( strlen( r->filename ) - 3 ), ".gz" ) == 0 ) {
			/**
			 * the user wanted the .gz content, so let's
			 * give it 'em
			 */
			return( DECLINED );
		}
	}

	if( r->filename[strlen( r->filename ) -1 ] == '/' ) {
		snprintf( scratch, size_of_scratch, "%sindex.html.gz",
			r->filename );
	}
	else {
		/**
		 * normal situation
		 */
		snprintf( scratch, size_of_scratch, "%s.gz", r->filename );
	}

	/*
	 * syslog( LOG_DEBUG, scratch );
	 */

	if( stat( scratch, &s ) ) {

		/**
		 * we cant find the file, lets leave it at that
		 * syslog( LOG_DEBUG, "we cannot find the file" );
		 */
		return( DECLINED );
	}

	FILE *f = fopen( scratch, "r" );

	if( !f ) {
		return( DECLINED );
	}

	/**
	 * now we need logic for the case where the user can accept
	 * gzip, they have the appropriate accept header
	 */

	set_mime_type( r, scratch );

	if( user_can_accept_gzip( r ) == EXIT_SUCCESS ) {
		/* ap_set_content_type( r, "text/html;charset=ascii" );
		 */
		apr_table_setn( r->headers_out, 
			"Content-Encoding", "gzip" );

		/*
		 * apr_table_setn( r->headers_out, 
			"Content-Type", "text/html" );
		*/


		while( (count = fread( scratch, 1, BUFFER, f ) ) ) {
			/*
			syslog( LOG_DEBUG, 
				"writing buffer to client %d", count );
			*/
			ap_rwrite( scratch, count, r );
		}
		fclose(f);
		return( DONE );
	}
	else {
		/* apr_table_setn( r->headers_out, 
			"Content-Type", "text/html" ); */
		gz = gzopen( scratch, "rb" );
		/*
		 * syslog( LOG_DEBUG, "Going to do a decompressed read" );
		 */
		while( ( count = gzread( gz, scratch, BUFFER ) ) > 0 ) {
			ap_rwrite( scratch, count, r );
		}
		gzclose( gz );
		return( DONE );
	}


	/* 
	 * syslog( LOG_DEBUG, "finished writing" );
	 */

	return( DECLINED );
}

/**
 * this is the business for the hook, it's a content generator, so it's
 * a default-style handler
 */
static void gzip_disk_hooks( apr_pool_t *pool ) {
	ap_hook_header_parser( gzip_disk_header_parser, NULL, NULL, APR_HOOK_MIDDLE );
}

/**
 * this is the definition for the module
 */
module AP_MODULE_DECLARE_DATA gzip_disk_module = {
	STANDARD20_MODULE_STUFF,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	gzip_disk_hooks
};

